#include <stdint.h>
#include <string>
#include <vector>
#include <sstream>
#include <map>
#include <set>
#include <csignal>
#include <cerrno>
#include <sys/stat.h>
#include <getopt.h>

#include <unistd.h>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/bind.hpp>

#include "hash.hh"
#include "sha1.hh"
#include "socket.hh"
#include "httpfun.hh"
#include "bdecode.hh"
#include "autoptr.hh"
#include "stringfun.hh"

#include <event.h> /* Using libevent for the polling facilities */

struct AutoCloser { AutoCloser(int s) : sock(s) { }
                    ~AutoCloser() { close(sock); }
                    int sock; };

struct ConnHandler: public ptrable
{
    virtual ~ConnHandler() { }
};

/******************/
/* manipulation of events (wrapped into Conn objects) */
/******************/
struct Conn
{
    int fd;
    event ev;
    std::string Received;
    std::string Sending;
    short flags;
    void (*handler)(int,short,void*);
    
    autoptr<ConnHandler> obj;
    
    Conn() : fd(0),ev(),Received(),Sending(),flags(0),handler(),obj() { }
};

static std::set<Conn*> ConnList;

static Conn* _SetupEvent(Conn* c, int fd,short flags,
    void(*handler)(int,short,void*),
    ConnHandler* obj)
{
    event_set(&c->ev, fd, flags, handler, (void*)c);
    event_add(&c->ev, 0);
    c->fd = fd;
    c->flags = flags;
    c->handler = handler;
    c->obj = obj;
    return c;
}
static Conn* UpdateEvent(Conn* c, short new_flags)
{
    if(new_flags == c->flags) return c;
    event_del(&c->ev);
    return _SetupEvent(c, c->fd, new_flags, c->handler, c->obj);
}
static Conn* CreateEvent(int fd, short flags, void (*handler)(int,short,void*),
    ConnHandler* obj = 0)
{
    Conn* c = new Conn;
    ConnList.insert(c);
    return _SetupEvent(c, fd,flags,handler, obj);
}
static void EndEvent(Conn* c)  
{
    close(c->fd);
    event_del(&c->ev);
    ConnList.erase(c);
    delete c;
}
/******************/

static const std::string LoadFileContents(const std::string& pathname)
{
    FILE* fp = fopen(pathname.c_str(), "rb");
    std::string result;
    if(fp)
    {
        for(;;)
        {
            char Buf[8192];
            int n = fread(Buf, 1, sizeof(Buf), fp);
            if(n == 0 && ferror(fp)) { result.clear(); return result; }
            if(n <= 0) break;
            result.append(Buf, Buf+n);
        }
        fclose(fp);
    }
    return result;
}

static bool is_local_ip(const std::string& s)
{
    if(s.compare(0,3,"10.") == 0) return true;
    if(s.compare(0,4,"127.") == 0) return true;
    if(s.compare(0,8,"192.168.") == 0) return true;
    if(s.compare(0,8,"169.254.") == 0) return true;
    if(s.compare(0,7,"172.16.") == 0) return true;
    if(s.compare(0,7,"172.17.") == 0) return true;
    if(s.compare(0,7,"172.18.") == 0) return true;
    if(s.compare(0,7,"172.19.") == 0) return true;
    if(s.compare(0,6,"172.2.") == 0) return true;
    if(s.compare(0,7,"172.30.") == 0) return true;
    if(s.compare(0,7,"172.31.") == 0) return true;
    return false;
}

static const std::string get_forwarded_ip(const std::string& ip,
    const std::map<std::string, std::string>& headers)
{
    std::map<std::string, std::string>::const_iterator i;
    i = headers.find("http_x_forwarded_for");
    if(i != headers.end())
    {
        const std::string& forw = i->second;
        std::string::size_type p = forw.find(',');
        if(p == forw.npos) return forw;
        std::string res = forw.substr(0, p);
        if(!is_local_ip(res)) return res;
        return forw.substr(p);
    }
    
    i = headers.find("http_client_ip");
    if(i != headers.end()) return i->second;

    i = headers.find("http_via");
    if(i != headers.end())
    {
        const std::string& via = i->second;
        std::string::size_type p = via.find(" for ");
        if(p != via.npos) return via.substr(p + 5);
    }
    
    i = headers.find("http_from");
    if(i != headers.end()) return i->second;
    
    return ip;
}

struct HTTPresponse // for throwing
{
    HTTPresponse(int cde, const std::string& ctype,
                          const std::string& msg)
        : code(cde),contenttype(ctype),content(msg),extraheaders()
    {
    }
    void AddHeader(const std::string& hdr, const std::string& value)
        { extraheaders[hdr]=value; }
    
    int code;
    std::string contenttype;
    std::string content;
    std::map<std::string, std::string> extraheaders;
    
    void Compress_Gzip()
    {
        //std::vector<unsigned char> result(content.size()
        
        // TODO
        // gzencode, 1
        //AddHeader("Content-Encoding", "gzip");
        //AddHeader("Vary",       "Accept-Encoding");
    }
    void Compress_Deflate()
    {
        // TODO
        // gzcompress 
        //AddHeader("Content-Encoding", "deflate");
        //AddHeader("Vary",       "Accept-Encoding");
    }
    
    const std::string ConvertToResponse() const
    {
        std::string resptype = "RESPONSE";
        switch(code)
        {
            case 200: resptype = "OK"; break;
            case 301:
            case 302: resptype = "ALMOST RIGHT"; break;
            case 400: resptype = "BAD REQUEST"; break;
            case 404: resptype = "NOT FOUND"; break;
            case 408: resptype = "CLUMSY FINGERS"; break;
            case 500: resptype = "OH THE PAIN"; break;
        }
        
        std::stringstream resptext;
        resptext << "HTTP/1.1 " << code << " " << resptype << "\r\n"
                    "Content-type: " << contenttype << "\r\n"
                    "Content-length: " << content.size() << "\r\n";
        for(std::map<std::string, std::string>::const_iterator
            i = extraheaders.begin();
            i != extraheaders.end();
            ++i)
        {
            resptext << i->first << ": " << i->second << "\r\n";
        }
        resptext << "\r\n" << content;
        return resptext.str();
    }
};

static unsigned timeout_downloaders_interval = 60*60;

#include "peer.hh"
#include "file.hh"

static hash_map<std::string /* info hash */, File> files;
static int listen_port = 6969;
static std::vector<std::string> accepted_dirs;
static std::string redirect_url = "http://tasvideos.org/";
static std::string dstate_file = "dstate";
static unsigned dstate_write_interval = 30;
static unsigned parse_dir_interval = 60;
static unsigned reannounce_interval = 30*60;
static unsigned response_size = 50;
static unsigned socket_timeout = 15;
static unsigned timeout_check_interval = 5;

struct StateConverter
{
    /* Routines for reading and writing the "dstate" file.
     * The dstate file is a bencoded structure consisting
     * of the list of allowed files, their download statistics
     * and the list of peers currently in the swarm.
     *
     * We could use some other mechanism, but currently we
     * just prefer to be compatible with official BitTorrent.
     */
    
    static void Load() // loads "dstate"
    {
        std::string Dstate = LoadFileContents(dstate_file);
        autoptr<CAtom> decoded = BDecode(Dstate);
        if(!decoded) return;
        const CAtomDict* info = decoded->get_dict();
        if(!info) return;
        
        time_t t = time(NULL);
        struct stat st;
        if(!stat(dstate_file.c_str(), &st))
        {
            t = st.st_mtime;
        }
        
        CAtomDict peers_dict     = info->getItemDict("peers");
        CAtomDict completed_dict = info->getItemDict("completed");
        CAtomDict allowed_dict   = info->getItemDict("allowed");

        const std::map<std::string, autoptr<CAtom> >& cdict = completed_dict.getValue();
        for(std::map<std::string, autoptr<CAtom> >::const_iterator
            i = cdict.begin(); i != cdict.end(); ++i)
        {
            if(!i->second) continue;
            const CAtomInt* val = i->second->get_int(); if(!val) continue;
            File& file = files[i->first];
            file.fclass    = File::Obsolete;
            file.infohash  = i->first;
            file.completed = val->getValue();
        }

        const std::map<std::string, autoptr<CAtom> >& adict = allowed_dict.getValue();
        for(std::map<std::string, autoptr<CAtom> >::const_iterator
            i = adict.begin(); i != adict.end(); ++i)
        {
            if(!i->second) continue;
            const CAtomDict* dic = i->second->get_dict(); if(!dic) continue;
            File& file = files[i->first];
            file.infohash  = i->first;
            file.name      = dic->getItemStr("name").getValue();
            file.file      = dic->getItemStr("file", file.name + ".torrent").getValue();
            file.path      = dic->getItemStr("path", file.file).getValue();
            file.numfiles  = dic->getItemInt("numfiles", 1).getValue();
            file.length    = dic->getItemInt("length", 0).getValue();
            file.fclass    = File::Allowed;
            
            CAtomDict peers_here = peers_dict.getItemDict(i->first);
            const std::map<std::string, autoptr<CAtom> >& pdict = peers_here.getValue();
            for(std::map<std::string, autoptr<CAtom> >::const_iterator
                p = pdict.begin(); p != pdict.end(); ++p)
            {
                if(!p->second) continue;
                const CAtomDict* pee = p->second->get_dict(); if(!pee) continue;
                Peer& peer = file.peers[p->first];
                peer.peer_id = p->first,
                peer.addr_ip   = pee->getItemStr("ip").getValue();
                peer.key       = pee->getItemStr("key").getValue();
                peer.left      = pee->getItemInt("left").getValue();
                // nat not used
                peer.addr_port = pee->getItemInt("port").getValue();
                peer.useragent = pee->getItemStr("useragents").getValue();
                peer.first_seen = t;
                peer.last_seen  = t;
            }
        }
    }

    static void Write() // writes "dstate"
    {
        autoptr<CAtomDict> peers_dict     = new CAtomDict;
        autoptr<CAtomDict> completed_dict = new CAtomDict;
        autoptr<CAtomDict> allowed_dict   = new CAtomDict;
        
        for(hash_map<std::string, File>::iterator
            i = files.begin(); i != files.end(); ++i)
        {
            const File& file = i->second;
            
            completed_dict->setItem(i->first, new CAtomInt(file.completed));
            if(file.fclass == File::Allowed)
            {
                autoptr<CAtomDict> allow_data = new CAtomDict;
                allow_data->setItem("name", new CAtomString(file.name));
                allow_data->setItem("file", new CAtomString(file.file));
                allow_data->setItem("path", new CAtomString(file.path));
                allow_data->setItem("numfiles", new CAtomInt(file.numfiles));
                allow_data->setItem("length",   new CAtomInt(file.length));
                allowed_dict->setItem(i->first, &*allow_data);
                
                autoptr<CAtomDict> peerlist = new CAtomDict;
                for(std::map<std::string, Peer>::const_iterator
                    j = file.peers.begin(); j != file.peers.end(); ++j)
                {
                    autoptr<CAtomDict> peer_data = new CAtomDict;
                    peer_data->setItem("ip",         new CAtomString(j->second.addr_ip));
                    peer_data->setItem("key",        new CAtomString(j->second.key));
                    peer_data->setItem("useragents", new CAtomString(j->second.useragent));
                    peer_data->setItem("left",       new CAtomInt(j->second.left));
                    peer_data->setItem("port",       new CAtomInt(j->second.addr_port));
                    peerlist->setItem(j->first, &*peer_data);
                }
                peers_dict->setItem(i->first, &*peerlist);
            }
        }

        CAtomDict result_dict;
        result_dict.setItem("peers",     &*peers_dict);
        result_dict.setItem("completed", &*completed_dict);
        result_dict.setItem("allowed",   &*allowed_dict);
        std::string dstate_new = result_dict.bencode();
        
        FILE* fp = fopen(dstate_file.c_str(), "wb");
        if(!fp) perror(dstate_file.c_str());
        else
        {
            fwrite(dstate_new.data(), 1, dstate_new.size(), fp);
            fclose(fp);
        }
    }
};

struct AllowedDirUpdater
{
private:
    static std::string UpdateOneFile(
        const std::string& pathname,
        const std::string& filename)
    {
        /* Read a torrent file, verify its validity, and gather
         * the relevant data into the Files struct.
         */
        
        try {
        std::string Torrent = LoadFileContents(pathname);
        autoptr<CAtom> decoded = BDecode(Torrent);
        if(const CAtomDict* dict = decoded->get_dict())
        {
            //fprintf(stderr, "dict: %s\n", dict->bencode().c_str());
            
            if(dict->getItemStr("announce").getValue().empty()
            && dict->getItemList("announce-list").empty()) throw __LINE__;
            
            CAtomDict infodict = dict->getItemDict("info");
            
            if(infodict.getItemStr("pieces").getValue().size() % 20) throw __LINE__;
            if(infodict.getItemInt("piece length").getValue() <= 0) throw __LINE__;
            const std::string name = infodict.getItemStr("name").getValue();
            if(name.empty() || name.find('/') != name.npos
                            || name.find('\\') != name.npos
                            || name[0]=='.'
                            || name[0]=='~') throw __LINE__;
            
            //fprintf(stderr, "has%s files\n", !infodict.getItem("files") ? "no" : "");
            //fprintf(stderr, "has%s length\n", !infodict.getItem("length") ? "no" : "");
            
            if(!infodict.getItem("files") == !infodict.getItem("length")) throw __LINE__;
            
            int_fast64_t length = infodict.getItemInt("length").getValue();
            if(length < 0) throw __LINE__;
            CAtomList infofiles = infodict.getItemList("files");
            for(unsigned a=0; a<infofiles.size(); ++a)
            {
                const CAtomDict* file = infofiles[a]->get_dict();
                if(!file) throw __LINE__;
                length += file->getItemInt("length").getValue();
            }
            
            std::string info_hash = SHA1(infodict.bencode());
            
            File& file = files[info_hash];
            
            file.infohash = info_hash;
            file.length = length;
            file.numfiles = infofiles.empty() ? 1 : infofiles.size();
            file.name = name;
            file.path = pathname;
            file.file = filename;
            file.torrentsize = Torrent.size();
            
            if(file.fclass != File::Allowed)
            {
                file.fclass = File::Allowed;
                /*
                fprintf(stderr, "Registered len(%u)num(%u) name(%s)path(%s)file(%s)\n",
                    (unsigned)file.length,
                    (unsigned)file.numfiles,
                    file.name.c_str(),
                    file.path.c_str(),
                    file.file.c_str()); */
            }
            return info_hash;
        }
        } catch(int l) { fprintf(stderr, "torrent rejected at line %d\n", l); }
        return std::string();
    }
public:
    static void Reload() // rereads the accepted dirs
    {
        std::set<std::string> found_infohashes;
        
        for(unsigned a=0; a<accepted_dirs.size(); ++a)
        {
            namespace fs = boost::filesystem;
            
            fs::path fpath(fs::system_complete(fs::path(accepted_dirs[a], fs::native)));
            if(!fs::exists(fpath)) continue;
            if(!fs::is_directory(fpath)) continue;
            fs::directory_iterator end_i;
            for(fs::directory_iterator dir_i(fpath); dir_i != end_i; ++dir_i)
                if(!is_directory(*dir_i))
                {
                    /*
                    fprintf(stderr, "s(%s)rn(%s)rd(%s)leaf(%s)\n",
                        dir_i->string().c_str(),
                        dir_i->root_name().c_str(),
                        dir_i->root_directory().c_str(),
                        dir_i->leaf().c_str());
                    */
                    std::string hash = UpdateOneFile(
                        dir_i->string(),
                        dir_i->leaf());
                    if(!hash.empty())
                    {
                        //fprintf(stderr, "ok '%s'\n", hash.c_str());
                        found_infohashes.insert(hash);
                    }
                }
        }
        
        // Now check the "files" list; for every "Allowed" hash that
        // was _not_ found in found_infohashes, set type as "Obsolete".
        for(hash_map<std::string, File>::iterator
            i = files.begin(); i != files.end(); ++i)
        {
            if(found_infohashes.find(i->first) == found_infohashes.end()
            && i->second.fclass == File::Allowed)
            {
                fprintf(stderr, "Obsoleting torrent %s (%s)\n",
                    i->second.name.c_str(),
                    i->first.c_str());
                i->second.fclass = File::Obsolete;
            }
        }
    }
};

struct RequestHandler: public ConnHandler
{
private:
    std::string ip;
    std::string request_line;
    std::string request_urn;
    std::string request_method;
    std::string request_proto;
    std::multimap<std::string, std::string> queryparams;
    std::map<std::string, std::string> headers;
    std::string user_agent;
    unsigned line_number;
    bool request_over;
    time_t last_read;
    unsigned request_number;
    
public:
    RequestHandler(const std::string& ipaddr)
        : ip(ipaddr),
          request_line(),
          request_urn(),
          request_method(),
          request_proto(),
          queryparams(),
          headers(),
          user_agent(),
          line_number(0),
          request_over(false),
          last_read(time(NULL)),
          request_number(0)
    {
    }
    bool IsDone() const
    {
        return request_over;
    }
    void CheckTimeout(Conn* c, time_t now)
    {
        if(!request_over && IsTimeout(now))
        {
            request_over = true;
            last_read    = now;
            line_number  = 0;
            
            request_proto = "HTTP/1.0";
            headers.clear();
            
            c->Sending += HTTPresponse(408, "text/plain; charset=utf-8", "you're too slow."
                                      ).ConvertToResponse();
            UpdateEvent(c, EV_READ | EV_WRITE | EV_PERSIST);
        }
    }
    bool IsTimeout(time_t now) const
    {
        return last_read + socket_timeout < now;
    }
    
    void TryParseRequest(Conn* c, unsigned new_context_start_pos)
    {
        last_read = time(NULL);
        while(!c->Received.empty())
        {
            /* If the newly added portion of the string includes a newline,
             * we have a line of command.
             */
            std::string::size_type pos = c->Received.find('\n', new_context_start_pos);
            if(pos == std::string::npos)
            {
                return;
            }
            
            std::string::size_type cmd_end = pos;
            
            /* Strip spaces from the end of the line for convenience.
             * This also takes care of '\r's.
             */
            while(cmd_end > 0 && isspace(c->Received[cmd_end-1])) --cmd_end;  
            
            std::string command
                (c->Received.begin(), c->Received.begin() + cmd_end);
            
            c->Received.erase(0, pos+1);
            
            ParseInputLine(c, command);
            
            new_context_start_pos = 0;
        }
    }
    
    void ParseInputLine(Conn* c, const std::string& line)
    {
        if(line_number == 0)
        {
            request_line = line;
            char method[100] = "", reqstr[100000] = "", proto[1000] = "";
            sscanf(line.c_str(), "%99s %99999s %999s", method, reqstr, proto);
            std::string req(reqstr);
            SplitDecodeQuery(req, queryparams);
            request_urn    = req;
            request_method = method;
            request_proto  = proto;
            
            if(request_proto != "HTTP/1.0"
            && request_proto != "HTTP/1.1")
            {
                request_over = true;
                c->Sending += HTTPresponse(400, "text/plain; charset=utf-8",
                    "Your browser sent a request we could not understand.")
                         .ConvertToResponse();
                UpdateEvent(c, EV_WRITE | EV_PERSIST);
                return;
            }
        }
        else
        {
            if(line.empty())
            {
                //if(request_number > 0)
                //printf("Got request %u (%s, %s, %s)\n", request_number,
                //  request_urn.c_str(), request_method.c_str(), request_proto.c_str());
                //fflush(stdout);
                
                user_agent = headers["user_agent"];
                
                // Done with input.
                HTTPresponse resp = HandleReq(get_forwarded_ip(ip, headers));

                //fprintf(stderr, "Got response: '%s' (%u bytes)\n", response.c_str(), response.size());
                
                char datetime[128];
                struct tm tm; { time_t t = time(NULL); localtime_r(&t, &tm); }
                strftime(datetime,sizeof(datetime), "%d/%b/%Y:%H:%M:%S", &tm);
                
                //if(false)
                printf("%s - - [%s] \"%s\" %d %lu \"-\" \"%s\"\n",
                    ip.c_str(),
                    datetime,
                    request_line.c_str(),
                    resp.code,
                    (unsigned long)resp.content.size(),
                    user_agent.c_str()
                );
                
                const std::string accept_encoding = headers["accept_encoding"];
                if(accept_encoding.find("gzip") != std::string::npos)
                    resp.Compress_Gzip();
                else if(accept_encoding.find("deflate") != std::string::npos)
                    resp.Compress_Deflate();

                c->Sending += resp.ConvertToResponse();
                
                bool keepalive = false;
                if(request_proto == "HTTP/1.1") keepalive = true;
                if(headers["connection"] == "close") keepalive = false;
                
                if(keepalive)
                {
                    ++request_number;
                    line_number = 0; // Back to beginning!
                }
                else
                {
                    request_over = true;
                }
                return;
            }
            
            const char* Buf = line.c_str();

            // It is a header.
            char* colon = strchr(Buf, ':');
            if(colon)
            {
                std::string hdrname = LCase(Trim(std::string(Buf, (const char*)colon)));
                for(unsigned a=0; a<hdrname.size(); ++a)
                    if(hdrname[a]=='-') hdrname[a]='_';
                while(*++colon == ' ');
                std::string hdrvalue= Trim(colon);
                headers[hdrname] = hdrvalue;
            }
        }
        ++line_number;
    }

private:    
    const std::string GetParam(const std::string& name, const std::string& def = "") const
    {
        std::multimap<std::string, std::string>::const_iterator
            i = queryparams.find(name);
        if(i == queryparams.end()) return def;
        return i->second;
    }
    
    File& FindFile(const std::string& info_hash)
    {
        hash_map<std::string, File>::iterator i = files.find(info_hash);
        if(i == files.end()) throw false;
        return i->second;
    }
    
private:
    const HTTPresponse HandleReq(const std::string& remote_ip)
    {
        try {
            if(request_method != "GET") throw HTTPresponse(400, "text/plain; charset=utf-8",
                "Your browser sent a request we could not understand.");
            
            if(request_urn == "/announce") { ReqAnnounce(remote_ip); }
            else if(request_urn == "/" || request_urn == "/index.html") { ReqIndex(); }
            else if(request_urn == "/scrape")   { ReqScrape(); }
            else if(request_urn == "/file")     { ReqFile(); }
            else { ReqOther(); }
        }
        catch(const HTTPresponse& resp)
        {
            return resp;
        }
        return HTTPresponse(500, "text/plain; charset=utf-8", "Internal server error.");
    }
    
    void ReqScrape() // req for torrent info
    {
        // A number of 'info_hash'es can be given.
        // If none is given, all allowed files should be listed.
        std::multimap<std::string, std::string>::const_iterator
            i = queryparams.find("info_hash");
        
        autoptr<CAtomDict> result = new CAtomDict;
        
        if(i == queryparams.end())
        {
            for(hash_map<std::string, File>::iterator
                i = files.begin(); i != files.end(); ++i)
            {
                if(i->second.fclass == File::Allowed)
                    ScrapeAddFile(*result, i->second);
            }
        }
        else do
        {
            try {
                ScrapeAddFile(*result, FindFile(i->second));
            } catch(bool) { }
        } while(++i != queryparams.end() && i->first == "info_hash");
        
        autoptr<CAtomDict> res2 = new CAtomDict;
        res2->setItem("files", &*result);
        
        throw HTTPresponse(200, "application/octet-stream", res2->bencode());
    }
    
    autoptr<CAtomDict> ScrapeCreateInfoDict(const File& file)
    {
        unsigned n_complete   = 0;
        unsigned n_incomplete = 0;
        
        for(std::map<std::string, Peer>::const_iterator
            i = file.peers.begin(); i != file.peers.end(); ++i)
        {
            if(i->second.left) ++n_incomplete; else ++n_complete;
        }
        
        autoptr<CAtomDict> filedata = new CAtomDict;
        filedata->setItem("downloaded", new CAtomInt(file.completed));
        filedata->setItem("incomplete", new CAtomInt(n_incomplete));
        filedata->setItem("complete",   new CAtomInt(n_complete));
        filedata->setItem("name", new CAtomString(file.name));
        return filedata;
    }
    
    void ScrapeAddFile(CAtomDict& target, const File& file)
    {
        autoptr<CAtomDict> filedata = ScrapeCreateInfoDict(file);
        target.setItem(file.infohash, new CAtomRawBencode(filedata->bencode()));
    }
    
    void ReqAnnounce( // req for peer info
        const std::string& remote_ip)
    {
        std::string inp_info_hash = GetParam("info_hash");
        std::string inp_peer_id   = GetParam("peer_id");
        std::string inp_key       = GetParam("key");
        std::string inp_ip        = remote_ip;
        long         inp_port     = strtol(GetParam("port", "0").c_str(), 0, 10);
        int_fast64_t inp_left     = strtoll(GetParam("left", "0").c_str(), 0, 10);
        long         numwant      = strtol(GetParam("numwant", "-1").c_str(), 0, 10);
        std::string event         = GetParam("event");
        
        int return_type = 0;
        
        if(GetParam("no_peer_id", "0") != "0") return_type = 1;
        if(GetParam("compact","0")     != "0") return_type = 2;
        
        if(numwant > response_size || numwant < 0) numwant = response_size;
        
        if(inp_info_hash.empty())   throw HTTPresponse(400, "text/plain; charset=utf-8", "needs info_hash");
        if(inp_peer_id.size() != 20)throw HTTPresponse(400, "text/plain; charset=utf-8", "id not of length 20");
        if(event != "started" && event != "completed"
        && event != "stopped" && event != "snooped"
        && event != "") 
            throw HTTPresponse(400, "text/plain; charset=utf-8", "invalid event");
        if(inp_port < 0 || inp_port > 65535) 
            throw HTTPresponse(400, "text/plain; charset=utf-8", "invalid port");
        if(inp_left < 0)
            throw HTTPresponse(400, "text/plain; charset=utf-8", "invalid amount left");
        
        try
        {
            File& file = FindFile(inp_info_hash);
            
            file.CleanupPeers();
            
            Peer& peer = file.peers[inp_peer_id];
            
            bool auth = peer.peer_id.empty()
                     || peer.key     == inp_key
                     || peer.addr_ip == inp_ip;

            if(event == "stopped")
            {
                if(auth)
                    file.peers.erase(inp_peer_id);
            }
            else if(peer.peer_id.empty())
            {
                peer.peer_id   = inp_peer_id;
                peer.addr_ip   = inp_ip;
                peer.addr_port = inp_port;
                peer.key       = inp_key;
                peer.left      = inp_left;
                peer.first_seen= time(NULL);
                peer.useragent = user_agent;
                
                if(event == "completed")
                    file.completed += 1;
            }
            else
            {
                if(auth) 
                {
                    // update stats
                    if(peer.left && !inp_left)
                        file.completed += 1;

                    if(peer.left) peer.left = inp_left;
                    peer.last_seen = time(NULL);
                }
            }

            autoptr<CAtomList> peerslist = new CAtomList;
            // each peers has "ip"(str), "peer id"(str), "port"(int)
            
            for(std::map<std::string, Peer>::const_iterator
                i = file.peers.begin(); i != file.peers.end(); ++i)
            {
                const Peer& peer = i->second;
                if(peer.peer_id == inp_peer_id) continue;
            
                switch(return_type)
                {
                    case 0: // default
                    {
                        autoptr<CAtomDict> peerdata = new CAtomDict;
                        peerdata->setItem("port",    new CAtomInt(peer.addr_port));
                        peerdata->setItem("ip",      new CAtomString(peer.addr_ip));
                        peerdata->setItem("peer id", new CAtomString(peer.peer_id));
                        peerslist->addItem(&*peerdata);
                        break;
                    }
                    case 1: // default, except peer id is not displayed
                    {
                        autoptr<CAtomDict> peerdata = new CAtomDict;
                        peerdata->setItem("port",    new CAtomInt(peer.addr_port));
                        peerdata->setItem("ip",      new CAtomString(peer.addr_ip));
                        peerslist->addItem(&*peerdata);
                        break;
                    }
                    case 2: // compact
                    {
                        std::string raw;
                        unsigned a=0,b=0,c=0,d=0;
                        sscanf(peer.addr_ip.c_str(), "%d.%d.%d.%d",&a,&b,&c,&d);
                        raw += (char)a;
                        raw += (char)b;
                        raw += (char)c;
                        raw += (char)d;
                        raw += (char)(peer.addr_port >> 8);
                        raw += (char)(peer.addr_port & 255);
                        peerslist->addItem(new CAtomRawBencode(raw));
                        break;
                    }
                }
            }
            
            peerslist->Randomize();
            peerslist->chop_to(numwant);
            
            if(return_type == 2) // compactify
            {
                std::string raw;
                for(unsigned a=0; a<peerslist->size(); ++a)
                    raw += (*peerslist)[a]->bencode();
                peerslist = (CAtomList*)(new CAtomString(raw));
            }
            
            autoptr<CAtomDict> result = ScrapeCreateInfoDict(file);
            // ^for complete, incomplete etc
            result->setItem("interval", new CAtomInt(reannounce_interval));
            result->setItem("peers",    &*peerslist);
            
            throw HTTPresponse(200, "application/octet-stream",
              result->bencode());
        }
        catch(bool)
        {
            throw HTTPresponse(200, "application/octet-stream",
                "d14:failure reason"
                 "63:Requested download is not authorized for use with this tracker."
                "e");
        }
    }
    
    void ReqFile() // req for a torrent file
    {
        // info_hash denotes the file to download
        try
        {
            std::string info_hash = GetParam("info_hash");
            File& file = FindFile(info_hash);
            
            HTTPresponse res(200, "application/x-bittorrent",
                LoadFileContents(file.path));
            res.AddHeader("Content-Disposition", "attachment; filename="+file.file);
            
            throw res;
        }
        catch(bool)
        {
            HTTPresponse res(404, "text/plain; charset=utf-8", "Your file is not known to this tracker.");
        }
    }
    
    void ReqIndex() // req for html torrent list
    {
        HTTPresponse res(302, "text/html",
            "Content is at <a href='"+htmlencode(redirect_url)+"'>"
           + htmlencode(redirect_url) + "</a>.");
        res.AddHeader("Location", redirect_url);
        res.AddHeader("Server",     "BQTracker-"VERSION);
        res.AddHeader("Connection", "close");
        throw res;
    }

    void ReqOther() // req for anything else; respond with an error
    {
        throw HTTPresponse(404, "text/plain; charset=utf-8", "Data not found at given address.");
    }
};

static void ConnectionCommunicator(int fd, short which, void* arg)
{
    Conn* c = (Conn*)arg;
    RequestHandler* r = (RequestHandler*) &* (c->obj);
    
    if(which & EV_READ)
    {
        unsigned had_len = c->Received.size();   
        for(;;)
        {
            char Buf[16384];
            int res = read(fd, Buf, sizeof(Buf));
            if(res > 0)
            {
                c->Received.append(Buf + 0, Buf + res);
                continue;  
            }
            if(res == 0) 
            {
                goto Close;
            }
            if(res == -1)
            {
                if(errno == EAGAIN || errno == EWOULDBLOCK) break;
                goto Close;
            }
        }
        
        r->TryParseRequest(c, had_len);
    }
    
    if(!c->Sending.empty())
    {
        if(!(c->flags & EV_WRITE)) which |= EV_WRITE; // try writing immediately
        /* While there is something to write, do not try to read new commands */
        UpdateEvent(c, EV_WRITE | EV_PERSIST);
    }

    if(which & EV_WRITE)
    {
        while(!c->Sending.empty())
        {
            const char* str = c->Sending.c_str();
            
            int res = write(fd, str, c->Sending.size());
            if(res > 0)
            {
                c->Sending.erase(0, res);
                continue;  
            }
            if(res == 0) 
            {
                goto Close;
            }
            if(res == -1)
            {
                if(errno == EAGAIN || errno == EWOULDBLOCK) break;
                goto Close;
            }
        }
    }
    
    if(c->Sending.empty())
    {
        if(r->IsDone()) goto Close;
        UpdateEvent(c, EV_READ | EV_PERSIST);
    }
    return;
Close:
    EndEvent(c);
}

static void CheckConnectionTimeouts()
{
    time_t t = time(NULL);
    
    for(std::set<Conn*>::iterator
        i = ConnList.begin(), j(i); i != ConnList.end(); i=j)
    {
        j = i; ++j;
        
        Conn* c = *i;
        if(c->obj)
        {
            RequestHandler* r = (RequestHandler*) &* (c->obj);
            r->CheckTimeout(c, t);
            if(r->IsTimeout(t))
            {
                EndEvent(c);
            }
        }
    }
}

static void ConnectionReceiver(int fd, short /*which*/, void* /*arg*/)
{
    //Conn* c = (Conn*)arg;
    for(;;)
    {
        int sfd = SockAcceptNonblock(fd);
        if(sfd < 0) break;
        CreateEvent(sfd, EV_READ | EV_PERSIST, ConnectionCommunicator,
            new RequestHandler(GetLastPeerIp())
                   );
    }
}

static void InitWriteStateLoop(int,short,void*)
{
    StateConverter().Write();
    
    static event tev;
    struct timeval tv = { dstate_write_interval, 0 };
    evtimer_set(&tev, InitWriteStateLoop, 0);
    evtimer_add(&tev, &tv);
}

static void InitReloaderLoop(int,short,void*)
{
    AllowedDirUpdater().Reload();
    
    static event tev;
    struct timeval tv = { parse_dir_interval, 0 };
    evtimer_set(&tev, InitReloaderLoop, 0);
    evtimer_add(&tev, &tv);
}

static void InitTimeoutCheckLoop(int,short,void*)
{
    CheckConnectionTimeouts();
    static event tev;
    struct timeval tv = { timeout_check_interval, 0 };
    evtimer_set(&tev, InitTimeoutCheckLoop, 0);
    evtimer_add(&tev, &tv);
}

static void MainLoop()
{
    event_init();

    InitWriteStateLoop(0,0,0);
    InitReloaderLoop(0,0,0);
    InitTimeoutCheckLoop(0,0,0);

    int plug = MakeServerSocket(listen_port);
    if(plug < 0) return;

    const AutoCloser AutoClose(plug);
    
    CreateEvent(plug, EV_READ | EV_PERSIST, ConnectionReceiver);

    event_dispatch();
}

int main(int argc, char** argv)
{
    std::signal(SIGPIPE, SIG_IGN);
    
    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"help",        0, 0,'h'},
            {"version",     0, 0,'V'},
            {"port",        1 ,0,'p'},
            {"dfile",       1, 0,'d'},
            {"save_dfile_interval", 1,0,500},
            {"timeout_downloaders_interval", 1,0,501},
            {"reannounce_interval", 1,0,502},
            {"response_size",       1,0,503},
            {"parse_dir_interval",  1,0,504},
            {"infopage_redirect",   1,0,505},
            {"allowed_dir",         1,0,506},
            {"socket_timeout",      1,0,507},
            {"timeout_check_interval", 1,0,508},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "hVp:d:", long_options, &option_index);
        if(c==-1) break;
        switch(c)
        {
            case 'V':
            {
                std::printf("%s\n", VERSION);
                return 0;
            }
            case 'h':
            {
                std::printf(
                    "bqtrack v"VERSION" - Copyright (C) 1992,2007 Bisqwit (http://iki.fi/bisqwit/)\n"
                    "\n"
                    "Tiny, fast BitTorrent tracker\n"
                    "\n"
                    "Usage: bqtrack [<options>]\n"
                    " --help, -h         This help\n"
                    " --version, -V      Displays version information\n"
                    " --port, -p <port>  Port to listen on. (defaults to 6969)\n"
                    " --dfile, -d <file> File to store recent downloader info in\n"
                    " --dsave_dfile_interval <nseconds>\n"
                    "                    Seconds between saving dfile (defaults to 30)\n"
                    " --timeout_downloaders_interval <nseconds>\n"
                    "                    Maximum age of a downloader's last contact\n"
                    "                    in seconds (defaults to 3600)\n"
                    " --reannounce_interval <nseconds>\n"
                    "                    Seconds downloaders should wait between\n"
                    "                    reannouncements (defaults to 1800)\n"
                    " --response_size <num>\n"
                    "                    Number of peers to send in an info message (defaults to 50)\n"
                    " --parse_dir_interval <nseconds>\n"
                    "                    Seconds between reloading of allowed_dirs (defaults to 60)\n"
                    " --socket_timeout <nseconds>\n"
                    "                    Maximum allowed wait in communication (defaults to 15)\n"
                    " --timeout_check_interval <nseconds>\n"
                    "                    Time to wait between checking if any connections have\n"
                    "                    timed out (defaults to 5)\n"
                    " --infopage_redirect <url>\n"
                    "                    URL to redirect to from index request\n"
                    " --allowed_dir <dir>\n"
                    "                    Adds the given directory to the list of allowed\n"
                    "                    directories (where allowed torrents are found)\n"
                );
                return 0;
            }
            case 'd':
                dstate_file = optarg;
                break;
            case 'p':
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1 || val > 65535)
                {
                    std::fprintf(stderr, "bqtrack: Invalid listen_port"); return 1;
                }
                listen_port = val;
                break;
            }
            case 500:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid save_dfile_interval"); return 1;
                }
                dstate_write_interval = val;
                break;
            }
            case 501:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid timeout_downloaders_interval"); return 1;
                }
                timeout_downloaders_interval = val;
                break;
            }
            case 502:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid reannounce_interval"); return 1;
                }
                reannounce_interval = val;
                break;
            }
            case 503:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid response_size"); return 1;
                }
                response_size = val;
                break;
            }
            case 504:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid parse_dir_interval"); return 1;
                }
                parse_dir_interval = val;
                break;
            }
            case 505:
                redirect_url = optarg;
                break;
            case 506:
                accepted_dirs.push_back(optarg);
                break;
            case 507:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid socket_timeout"); return 1;
                }
                socket_timeout = val;
                break;
            }
            case 508:
            {
                char* arg = optarg;
                long val = strtol(optarg, &arg, 10);
                if(val < 1)
                {
                    std::fprintf(stderr, "bqtrack: Invalid timeout_check_interval"); return 1;
                }
                timeout_check_interval = val;
                break;
            }
        }
    }
    if(argc != optind)
    {
        std::fprintf(stderr, "bqtrack: invalid parameters. See `bqtrack --help'\n");
        return 1;
    }
    
    if(accepted_dirs.empty())
    {
        std::fprintf(stderr, "bqtrack: Refusing to start tracking with no allowed files. See `bqtrack --help'\n");
        return 1;
    }

    StateConverter().Load();
    MainLoop();
    
    return 0;
}
